import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.Vector;

public class ConflictAnalyzer {
	BufferedWriter bw;
	Iterator<Integer> it_int;
	Iterator<Actions> it_act;
	Iterator<String> it;
	YicesModelFileParser parser;
	static final File yices_dir=new File("/home/astridr/tools/yices-2.6.1/bin");
	IdManager im;
	YicesWriter writer;
	YicesModelFileParser yp;
	String data_dir;
	Runtime rt;
	
	public ConflictAnalyzer(){
		this.im=new IdManager();
	
		writer=new YicesWriter();	
		writer=new YicesWriter(this.im);
		yp=new YicesModelFileParser();
		rt=Runtime.getRuntime();
		data_dir="/home/astridr/programming/java/YicesTools/data/";
	}
	
	public static void main(String[] args){
		try{
		ConflictAnalyzer ca=new ConflictAnalyzer();
		ca.determineMaximalGoalSets();
		}catch(Exception exc){
			exc.printStackTrace();
		}
	}
	
	public void determineMaximalGoalSets()throws Exception{
		yp.parse(data_dir+"conflict3.ys.txt");
		YicesObject yo=new YicesObject();
		yo.becomeCrossProductActionsObject(yp.yo);
		
		boolean res1;
		boolean res2;
		it_int=yo.goals_at_prio.keySet().iterator();
		Vector<String> all_goals=new Vector<String>();
		while(it_int.hasNext()){
			all_goals.addAll(yo.goals_at_prio.get(it_int.next()));
		}
		Vector<Vector<String>> goal_sets=getGoalSets(all_goals);
		boolean found_winning=false;
		
		YicesObject temp_yo;
		for(int iii=0;iii<goal_sets.size();iii++){
			System.out.println("Check for WS for "+goal_sets.get(iii).size()+" "+goal_sets.get(iii));
			temp_yo=new YicesObject(yo);
			temp_yo.setGoals(goal_sets.get(iii));
			writer.write(data_dir+"conflict.ys",yo);	
			res1=this.executeYices(rt,data_dir+"conflict.ys",data_dir+"conflict.res");
			if(res1){
				writer.writeNegGoal("data/conflict_neg.ys",temp_yo);
				res2=this.executeYices(rt,data_dir+"conflict_neg.ys",data_dir+"conflict_neg.res");
				if(res2){
					found_winning=checkWinning(1,temp_yo,"_"+this.im.getGoalSetId(goal_sets.get(iii)));	
				}else{
					found_winning=true;
				}
			}
			if(found_winning){
				System.out.println("Found WS for "+goal_sets.get(iii).size()+" "+goal_sets.get(iii));
			}else{
				System.out.println("Found NOWS for "+goal_sets.get(iii).size()+" "+goal_sets.get(iii));
				//determine Missing Information
				Infos info_sets=getInfoSets(yo.info_list);
				for(int jjj=0;jjj<info_sets.size();jjj++){
					System.out.println("Check for WS for "+goal_sets.get(iii).size()+" "+goal_sets.get(iii));
					temp_yo=new YicesObject(yo);
					temp_yo.setGoals(goal_sets.get(iii));
					temp_yo.conjunction_list.add(info_sets.get(jjj));
					writer.write(data_dir+"conflict_i"+jjj+".ys",yo);	
					res1=this.executeYices(rt,data_dir+"conflict_i"+jjj+".ys",data_dir+"conflict.res");
					if(res1){
						writer.writeNegGoal("data/conflict_neg_i"+jjj+".ys",temp_yo);
						res2=this.executeYices(rt,data_dir+"conflict_neg_i"+jjj+".ys",data_dir+"conflict_neg.res");
						if(res2){
							found_winning=checkWinning(1,temp_yo,"_"+this.im.getGoalSetId(goal_sets.get(iii)));	
						}else{
							found_winning=true;
						}
					}
					if(found_winning){
						System.out.println("Found WS for "+goal_sets.get(iii).size()+" "+goal_sets.get(iii)+" under "+info_sets.get(jjj));
					}else{
						System.out.println("No winning strategy");
					}
				}
			}
		}
	}
	
	public Infos buildDisjunction(Vector<Infos> info_sets)throws Exception{
		Infos infos;
		Infos new_infos;
		Vector<Infos> new_info_sets=new Vector<Infos>();
		StringBuffer sb=new StringBuffer();
		new_infos=new Infos();
		for(int iii=0;iii<info_sets.size();iii++){
			infos=info_sets.get(iii);

			if(infos.size()<1){throw new Exception("Empty info");}
			String info=infos.get(0);
			sb=new StringBuffer();
			for(int jjj=1; jjj<infos.size();jjj++){
				sb.append("(or ");
				sb.append(info);
				sb.append(" ");
				sb.append(infos.get(jjj));
				sb.append(")");
				info=sb.toString();
			}
			new_infos.add(info);
		}
		//System.out.println("OI"+info_sets);
		//System.out.println("NI"+new_infos);
		return new_infos;
	}
	
	public Infos buildConjunction(int start,Vector<Infos> info_set){
		Infos res_infos=new Infos();
		Infos infos;
		if(info_set.size()<2){
			if(info_set.size()==0){
				return new Infos();
			}
			return info_set.get(0);
		}
		infos=info_set.get(start);
		if(start==info_set.size()-1){
			return infos;
		}
		Infos infos2=buildConjunction(start+1,info_set);
		StringBuffer sb;
		for(int iii=0;iii<infos.size();iii++){
			for(int kkk=0; kkk<infos2.size();kkk++){
				sb=new StringBuffer();
				sb.append("(and ");
				sb.append(infos.get(iii));
				sb.append(" ");
				sb.append(infos2.get(kkk));
				sb.append(")");
				res_infos.add(sb.toString());
				sb=new StringBuffer();
			}
		}
		System.out.println("Res "+res_infos+" "+res_infos.size());
		return res_infos;
	}
	/**
	 * determines all subset sets
	 * @param all_goals
	 * @return
	 */
	public Infos getInfoSets(Vector<Infos> all_infos)throws Exception{
		Infos tmp_res=new Infos();
		if(all_infos.size()==0){return tmp_res;}
		Vector<Infos> info_set=new Vector<Infos>();
		Infos infos;
		for(int iii=0; iii<all_infos.size();iii++){
				Infos tmp_info=all_infos.get(iii);
				Vector<Infos> info_sets=tmp_info.getSubsetSet();
				System.out.println("INFOS "+info_sets);
				infos=buildDisjunction(info_sets);
				info_set.add(infos);
		}
		return buildConjunction(0,info_set);
	}
	/**
	 * determines all subset sets
	 * @param all_goals
	 * @return
	 */
	public Vector<Vector<String>> getGoalSets(Vector<String> all_goals){
		Vector<Vector<String>> goal_sets=new Vector<Vector<String>>();
		goal_sets.add(all_goals);
		if(all_goals.size()>1){
			for(int iii=0; iii<all_goals.size();iii++){
				Vector<String> tmp_gs=new Vector<String>();
				tmp_gs.addAll(all_goals);
				tmp_gs.remove(all_goals.get(iii));
				goal_sets.addAll(getGoalSets(tmp_gs));
			}
		}
		return goal_sets;
	}
	/**
	 * depth first search over all actions of all rounds, i.e., fixes action for current round and then examines actions of the next round 
	 * @param round
	 * @param yo
	 * @param act_trace
	 * @throws Exception
	 */
	public boolean checkWinning(int round, YicesObject yo, String act_trace) throws Exception{
		//System.out.println("Actions at Round "+yo.actions_at_round+" || "+yo.actions_at_round.get(round));
		//if(round<=yo.actions_at_round.keySet().size()){
		//	System.out.println("Check Winning "+round+" "+act_trace);
		//}
		if(round>yo.actions_at_round.keySet().size()){
			if(round<=yo.actions_at_round.keySet().size()){
				//System.out.println("Reached maximal number of rounds of "+yo.actions_at_round.keySet().size());c
			}
			return false;
		}
		//it_act=yo.actions_at_round.get(round).iterator();
		//exculde one action if there are alternatives
		String action;
		String action_id;
		if(yo.actions_at_round.get(round).size()!=1){throw new Exception("Unexpected numer of actions. Expected crossproduct.");}
		//while(it_act.hasNext()){
		Actions act=yo.actions_at_round.get(round).iterator().next();
		//System.out.println("Actions "+act);
		it=act.iterator();
		int iteration=0;
		while(iteration<act.size()){
			//action=it.next();
			action=act.get(iteration);
			iteration++;
			action_id=act_trace+"+"+this.im.getActId(action).toString();
			YicesObject yo_tmp=new YicesObject(yo);
			yo_tmp.setAction(action, round);
			//System.out.println("Only Action "+action+" "+action_id+" in round "+round);
			//yo_tmp.printYicesObject();
			writer.write(data_dir+"conflict"+action_id+".ys",yo_tmp);	
			//          System.out.println("Goals "+yo_tmp.goals_at_prio);
			boolean res1=this.executeYices(rt,data_dir+"conflict"+action_id+".ys",data_dir+"conflict.res");
			if(res1){
				writer.writeNegGoal("data/conflict"+action_id+"_neg.ys",yo_tmp);
				boolean res2=this.executeYices(rt,data_dir+"conflict"+action_id+"_neg.ys",data_dir+"conflict_neg.res");
				if(res2){
					res2=checkWinning(round+1,yo_tmp,action_id);
					if(res2){
						return true;
					}
				}else{
					System.out.println("Found winning strategy "+data_dir+"conflict"+action_id+".ys");
					return true;
				}
			}
			//System.out.println("|"+act_trace+" it "+iteration+" "+it.hasNext()+" "+act.size());
		}
		return false;
		//}
	}
	
	
	public boolean executeYices(Runtime rt,String infile, String outfile)throws Exception{
		String command="/home/astridr/tools/yices-2.6.1/bin/yices "+infile;
		//TODO System.out.println("CMD "+command);
		Process p=rt.exec("/home/astridr/tools/yices-2.6.1/bin/yices "+infile,null,null);
		p.waitFor();
		BufferedReader br=new BufferedReader(new InputStreamReader(p.getInputStream()));
		BufferedReader br_error=new BufferedReader(new InputStreamReader(p.getErrorStream()));
		String line;
		BufferedWriter yices_result_bw=new BufferedWriter(new FileWriter(outfile));
		boolean first=true;
		String is_sat="";
		while((line=br.readLine())!=null){
			if(first){
				first=false;
				is_sat=line.trim();
			}
			yices_result_bw.write(line);
		}
		yices_result_bw.flush();
		yices_result_bw.close();
		int exit_val=p.exitValue();
		if(exit_val!=0){
			StringBuffer sb=new StringBuffer();
			while((line=br_error.readLine())!=null){
				sb.append(line);
			}
			throw new Exception("Exit value is "+exit_val+" "+sb.toString());
		}
		//TODO System.out.println("SAT "+is_sat);
		if(is_sat.equals("sat")){return true;}
		if(is_sat.equals("unsat")){return false;}
		throw new Exception("Unexpected format "+is_sat);
	}

}
