import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Vector;


public class YicesModelFileParser {
	boolean skip=false;
	public static final String ACTION_ROUND="<action round=";
	public static final String INFO="<info";
	public static final String GOAL_PRIO="<goal priority=";

	public YicesObject yo;
	
	BufferedReader br;
	int line_number;
	String line="";
	String parse_file="";
	
	public static void main(String args[]) {
		try{
			YicesModelFileParser parser=new YicesModelFileParser();
			parser.parse("/home/astridr/programming/java/YicesTools/data/conflict3.ys.txt");
		}catch(Exception exc){
			exc.printStackTrace();
		}
	}
	
	
	public void parse(String file)throws Exception{
		line_number=0;
		if( ! new File(file).exists() ){ throw new Exception("No system file"+file);}
		br=new BufferedReader(new FileReader(file));
		this.parse_file=file;
		while( skip || (line=this.readLine())!=null){
			line=line.trim();
			if(line.startsWith("//")){
				continue;
			}
			if(line.startsWith("<define>")){
				this.parseDefine();
			}
			if(line.startsWith("<conjunctions>")){
				this.parseConjunctions();
			}
			if(line.startsWith("<actions>")){
				this.parseActions();
			}
			if(line.startsWith("<infos>")){
				this.parseInfos();
			}
			if(line.startsWith("<goals>")){
				this.parseGoals();
			}
			if(line.startsWith("<mc>")){
				this.parseMC();
			}
			//System.out.println("L"+line);
			if(!(line.startsWith("<define>")||line.startsWith("<conjunctions>")||line.startsWith("<actions>")||
					line.startsWith("<infos>")||line.startsWith("<goals>")||line.startsWith("<mc>"))){
				throw new Exception("Unknown format of model file."+line);
			}
		}
		//TODO System.out.println("Print parsed Object:");
		//TODO this.yo.printYicesObject();
	}
	
	public void parseDefine()throws Exception{
		//System.out.println("parseSystem");
		this.yo=new YicesObject();
		this.yo.define_list=new Vector<String>();
		String line="";
		while((line=this.readLine())!=null){
			line=line.trim();
			if(line.endsWith("</define>")){
				return;
			}
			this.yo.define_list.add(line);
			
		}
	}
	
	public void parseConjunctions()throws Exception{
		//System.out.println("parseSystem");
		this.yo.conjunction_list=new Vector<String>();
		String line="";
		while((line=this.readLine())!=null){
			line=line.trim();
			if(line.endsWith("</conjunctions>")){
				return;
			}
			this.yo.conjunction_list.add(line);
			
		}
	}
	
	public void parseInfos()throws Exception{
	//	System.out.println("parseInfos");
		String line="";
		this.yo.info_list=new Vector<Infos>();
		while((line=this.readLine())!=null){
			line=line.trim();
			if(line.endsWith("</infos>")){
				return;
			}
			parseInfoLine(line);
		}
	}
	
	
	public void parseInfoLine(String line)throws Exception{
		//System.out.println("Action Line "+line);
		if(!line.startsWith(INFO)){throw new Exception("Unknown format of "+line);}
		String infos=line.substring(line.indexOf(">"),line.indexOf("</info>"));
		String info;
		Infos infos_c=new Infos();
		while(infos.length()>0){
			info=infos.substring(infos.indexOf("<alt>")+5,infos.indexOf("</alt>"));
			infos_c.add(info);
			infos=infos.substring(infos.indexOf("</alt>")+6).trim();
		}
		this.yo.info_list.add(infos_c);		
	}
	
	
	public void parseActions()throws Exception{
		//System.out.println("parseActions");
		String line="";
		this.yo.actions_at_round=new ActionsAtRound();
		while((line=this.readLine())!=null){
			line=line.trim();
			if(line.endsWith("</actions>")){
				return;
			}
			parseActionLine(line);
			
		}
	}
	public void parseActionLine(String line)throws Exception{
		//System.out.println("Action Line "+line);
		if(!line.startsWith(ACTION_ROUND)){throw new Exception("Unknown format of "+line);}
		String round=line.substring(ACTION_ROUND.length(),line.indexOf(">"));
		Integer round_int=Integer.parseInt(round);
		String actions=line.substring(line.indexOf(">"),line.indexOf("</action"));
		//System.out.println("Actions"+actions);
		String action;
		Actions actions_r=new Actions();
		while(actions.length()>0){
			action=actions.substring(actions.indexOf("<alt>")+5,actions.indexOf("</alt>"));
			//System.out.println("Action "+action);
			actions_r.add(action);
			actions=actions.substring(actions.indexOf("</alt>")+6).trim();	
		}
		this.yo.actions_at_round.addRoundActions(round_int, actions_r);		
	}
	
	
	public void parseGoals()throws Exception{
		//System.out.println("parseGoals");
		String line="";
		this.yo.goals_at_prio=new GoalsAtPriority();
		while((line=this.readLine())!=null){
			line=line.trim();
			//System.out.println("G "+line);
			if(line.endsWith("</goals>")){
				return;
			}
			parseGoalLine(line);
			
		}
	}
	public void parseGoalLine(String line)throws Exception{
		//System.out.println("Goal Line "+line);
		if(!line.startsWith(GOAL_PRIO)){throw new Exception("Unknown format of "+line);}
		String prio=line.substring(GOAL_PRIO.length(),line.indexOf(">"));
		Integer prio_int=Integer.parseInt(prio);
		String goal=line.substring(line.indexOf(">")+1,line.indexOf("</goal"));
		//System.out.println("Goals"+goal);
		this.yo.goals_at_prio.addPriorityGoals(prio_int, goal);		
	}
	
	public void parseMC()throws Exception{
		//System.out.println("parseSystem");
		this.yo.mc_list=new Vector<String>();
		String line="";
		while((line=this.readLine())!=null){
			line=line.trim();
			if(line.endsWith("</mc>")){
				return;
			}
			this.yo.mc_list.add(line);
		}
	}

	protected String readLine()throws Exception{
		this.line_number++;
		String line=this.br.readLine();
		//System.out.println("l "+line);
		line=removeComment(line);
		return line;
	}

	private String removeComment(String line){
		if(line==null){return line;}
		int idx=line.indexOf("//");
		if(idx==0){
			return "";
		}else if(idx>0){
			return line.substring(0,idx);
		}
		return line;
	}
}
